{==EZMRCI=============================================================

A unit to access the Microsoft Realtime Compression Interface. The
MRCI API is the compression/decompression engine as used by
DoubleSpace, and this unit enables the compression and decompression
of user defined blocks using the MRCI engine. This unit requires an
MRCI server to be present (and at the time of writing, this means
DoubleSpace); it does not implement MRCI in code.

Restrictions: the MRCI incremental decompression feature has not been
implemented; in protected mode (DOS and Windows) the maximum buffer
size for (de)compression is 8192 bytes.

EZMRCI is Copyright (c) 1994 by  Julian M. Bucknall

VERSION HISTORY
08Feb94 JMB 1.00 initial release
======================================================================}

unit EzMRCI;

{------Common compiler switches---------------------------------------}
{$A+   Word align variables }
{$B-   Short-circuit boolean expressions }
{$F+   Force Far calls }
{$I-   No I/O checking }
{$N+   Allow coprocessor instructions }
{$P+   Open parameters enabled }
{$Q-   No integer overflow checking }
{$R-   No range checking }
{$S-   No stack checking }
{$T-   @ operator is NOT typed }
{$V-   Disable var string checking }
{$X+   Enable extended syntax }
{$IFDEF DEBUG}
{$D+,L+,Y+  Enable debug information }
{$ENDIF}
{---------------------------------------------------------------------}

{------Real mode compiler switches------------------------------------}
{$IFDEF MSDOS}
{$E+   Enable coprocessor emulation }
{$G-   8086 type instructions }
{$O-   Do NOT allow overlays }
{$DEFINE RealMode}
{$UNDEF  ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

{------Protected mode compiler switches-------------------------------}
{$IFDEF DPMI}
{$E+   Enable coprocessor emulation }
{$G+   80286+ type instructions }
{$UNDEF  RealMode}
{$DEFINE ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

{------Windows compiler switches--------------------------------------}
{$IFDEF WINDOWS}
{$G+   80286+ type instructions }
{$K+   Use smart callbacks
{$W-   No Windows realmode stack frame }
{$UNDEF  RealMode}
{$DEFINE ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

INTERFACE

uses
  {$IFDEF MSDOS}
  Dos;
  {$ELSE}
  EzDPMI,                      {for protected mode services}
  WinDOS;
  {$ENDIF}

const
  {MRCI error codes}
  mrciErrSuccess        = 0;   {no error}
  mrciErrNotSupported   = 1;   {operation not supported}
  mrciErrBusy           = 2;   {MRCI server is busy, try again later}
  mrciErrBufferOverflow = 3;   {destination buffer is too small}
  mrciErrCannotCompress = 4;   {source buffer is incompressible}
  mrciErrBadFormat      = 5;   {source buffer is not in MRCI format}
  mrciErrBadBufferSize  = 99;  {Buffer is too large to (de)compress}

  {MRCI capabilities}
  mrciCapNone       = 0;       {No capabilities}
  mrciCapCompress   = $0001;   {Can compress}
  mrciCapDecompress = $0002;   {Can decompress}
  mrciCapIncDecomp  = $0020;   {Can decompress incrementally}

type
  PMRCIInfo = ^TMRCIInfo;
  TMRCIInfo = record
    miVendor        : array [1..4] of char; {Vendor ID}
    miServerVersion : word;    {MRCI server version}
    miMRCIVersion   : word;    {MRCI version}
    miEntryPoint    : pointer; {MRCI (de)compress realmode entry point}
    miCapability    : word;    {Server capabilities}
    miHWAssisted    : word;    {Hardware assisted capabilities}
    miMaxBuffer     : word;    {Max buffer size, will be >= 8192}
  end;

{=MRCIPresent=========================================================
Returns true if an MRCI driver was found.
08Feb94 JMB
======================================================================}
function MRCIPresent : boolean;

{=GetMRCIInfo=========================================================
Returns information about an MRCI server. miVendor is the server's
vendor ID (Microsoft use 'MSFT'). miServerVersion is the version
number of the server (hi byte = major, lo byte = minor). miMRCIVersion
is the version number of the MRCI spec supported (hi byte = major, lo
byte = minor). miEntryPoint is the entry point of the MRCI server in
realmode segment/offset form. miCapability is the set of capabilities
of the server and consists of mrciCapXxxx bit settings. miHWAssisted
is the set of capabilities performed by hardware in the system.
miMaxBuffer is the maximum buffer size that can be handled by the MRCI
server. The function result is one of the mrciErrXxxx codes.
08Feb94 JMB
======================================================================}
function GetMRCIInfo(var Info : TMRCIInfo) : word;

{=MRCICompress========================================================
Compresses a buffer containing uncompressed data into another buffer.
The buffers cannot be the same memory block. NormalSize is the number
of bytes to compress, and CompressSize is the number of bytes of
resulting compressed data if the function returns mrciErrSuccess. The
function result is one of the mrciErrXxxx codes.

WARNING: the MRCI spec warns that CompressBuffer must be at least as
large as NormalBuffer. Tests have shown that CompressBuffer should be
at least 1024 bytes in size. Hence, internally, MRCICompress *assumes*
that the size of CompressBuffer is the larger of 1024 or NormalSize.
08Feb94 JMB
======================================================================}
function MRCICompress(var NormalBuffer;
                      var CompressBuffer;
                          NormalSize : word;
                      var CompressSize : word) : word;

{=MRCIDecompress======================================================
Decompresses a buffer containing compressed data (CompressBuffer) into
another buffer (NormalBuffer). The buffers cannot be the same memory
block. CompressSize is the number of bytes of compressed data. On
input NormalSize is the expected size of the uncompressed data in
bytes, and on return (if the function result is mrciErrSuccess) is the
actual number of bytes. The function result is one of the mrciErrXxxx
codes.

WARNING: it is *your* responsibility to ensure that NormalBuffer is
large enough to hold the uncompressed data, MRCI does no checking
whatsoever. Also the MRCI decompressor will use the expected number of
bytes (input to the routine in NormalSize) to know when to stop
decompressing.
08Feb94 JMB
======================================================================}
function MRCIDecompress(var NormalBuffer;
                        var CompressBuffer;
                        var NormalSize : word;
                            CompressSize : word) : word;

IMPLEMENTATION

const
  MinCompressDestSize = 1024;   {on compression, min size of dest buffer}
  MaxDPMIBufferSize   = 8192;   {max buffer size in protected mode}

type
  OS = record O, S : word; end; {to decompose pointers into seg/sel & ofs}

type
  TMRCIRequest = record
    mrSrc         : pointer;    {pointer to Source buffer}
    mrSrcSize     : word;       {size of Source buffer in bytes}
    mrReserved    : word;
    mrDest        : pointer;    {pointer to Destination buffer}
    mrDestSize    : word;       {size of destination buffer}
    mrGranularity : word;       {granularity}
    mrIncrState   : longint;    {incremental decompression save state}
  end;

{$IFDEF ProtMode}
type
  PDPMIBuffer = ^TDPMIBuffer;
  TDPMIBuffer = record
    Normal  : array [0..pred(MaxDPMIBufferSize)] of byte;
    Comprss : array [0..pred(MaxDPMIBufferSize)] of byte;
    Req     : TMRCIRequest;
  end;
{$ENDIF}

var
  MRCIInfo : PMRCIInfo;         {pointer to MRCI server's info block}
  MRCIIsThere : boolean;        {true if MRCI detected}
  InDOSFlag : ^byte;            {pointer to InDOS flag}
  {$IFDEF ProtMode}
  ExitSave : pointer;           {saved ExitProc address}
  MRCIInfoReal : pointer;       {realmode pointer to MRCI info block}
  RealBuffer : PDPMIBuffer;     {realmode pointer to buffers}
  ProtBuffer : PDPMIBuffer;     {protected mode pointer to same buffers}
  {$ENDIF}

function MRCIPresent : boolean;
  begin
    MRCIPresent := MRCIIsThere;
  end;

function GetMRCIInfo(var Info : TMRCIInfo) : word;
  begin
    if MRCIIsThere then
      begin
        Info := MRCIInfo^;
        {$IFDEF ProtMode}
        Info.miMaxBuffer := MaxDPMIBufferSize;
        {$ENDIF}
        GetMRCIInfo := mrciErrSuccess;
      end
    else
      GetMRCIInfo := mrciErrNotSupported;
  end;

{$IFDEF ProtMode}
function MRCICompress(var NormalBuffer;
                      var CompressBuffer;
                          NormalSize : word;
                      var CompressSize : word) : word;
  var
    Regs : TRegisters;
  begin
    if not MRCIIsThere then
      MRCICompress := mrciErrNotSupported
    else if (NormalSize > MaxDPMIBufferSize) then
      MRCICompress := mrciErrBadBufferSize
    else if (InDOSFlag^ <> 0) then
      MRCICompress := mrciErrBusy
    else
      begin
        FillChar(ProtBuffer^.Req, sizeof(TMRCIRequest), 0);
        with ProtBuffer^.Req do
          begin
            mrSrc := RealBuffer;
            mrSrcSize := NormalSize;
            mrDest := RealBuffer;
            inc(OS(mrDest).O, MaxDPMIBufferSize);
            if (NormalSize < MinCompressDestSize) then
                 mrDestSize := MinCompressDestSize
            else mrDestSize := NormalSize;
            mrGranularity := 1; {ie maximum compression}
          end;
        Move(NormalBuffer, ProtBuffer^.Normal, NormalSize);
        FillChar(Regs, sizeof(Regs), 0);
        with Regs do
          begin
            ax := mrciCapCompress;
            es := OS(MRCIInfoReal).S;
            bx := OS(MRCIInfoReal).O;
            ds := OS(RealBuffer).S;
            si := OS(RealBuffer).O + 2*MaxDPMIBufferSize;
          end;
        asm
          mov ax, $8001
          int $2A
        end;
        RealCall(RealProc(MRCIInfo^.miEntryPoint), Regs);
        asm
          mov ax, $8101
          int $2A
        end;
        if (Regs.ax = 0) then
          begin
            CompressSize := ProtBuffer^.Req.mrDestSize;
            Move(ProtBuffer^.Comprss, CompressBuffer, CompressSize);
            MRCICompress := mrciErrSuccess;
          end
        else
          MRCICompress := Regs.ax;
      end;
  end;

function MRCIDecompress(var NormalBuffer;
                        var CompressBuffer;
                        var NormalSize : word;
                            CompressSize : word) : word;
  var
    Regs : TRegisters;
  begin
    if not MRCIIsThere then
      MRCIDecompress := mrciErrNotSupported
    else if (NormalSize > MaxDPMIBufferSize) then
      MRCIDecompress := mrciErrBadBufferSize
    else if (InDOSFlag^ <> 0) then
      MRCIDecompress := mrciErrBusy
    else
      begin
        FillChar(ProtBuffer^.Req, sizeof(TMRCIRequest), 0);
        with ProtBuffer^.Req do
          begin
            mrSrc := RealBuffer;
            inc(OS(mrSrc).O, MaxDPMIBufferSize);
            mrDest := RealBuffer;
            mrDestSize := NormalSize;
          end;
        Move(CompressBuffer, ProtBuffer^.Comprss, CompressSize);
        FillChar(Regs, sizeof(Regs), 0);
        with Regs do
          begin
            ax := mrciCapDecompress;
            es := OS(MRCIInfoReal).S;
            bx := OS(MRCIInfoReal).O;
            ds := OS(RealBuffer).S;
            si := OS(RealBuffer).O + 2*MaxDPMIBufferSize;
          end;
        asm
          mov ax, $8001
          int $2A
        end;
        RealCall(RealProc(MRCIInfo^.miEntryPoint), Regs);
        asm
          mov ax, $8101
          int $2A
        end;
        if (Regs.ax = 0) then
          begin
            NormalSize := ProtBuffer^.Req.mrDestSize;
            Move(ProtBuffer^.Normal, NormalBuffer, NormalSize);
            MRCIDecompress := mrciErrSuccess;
          end
        else
          MRCIDecompress := Regs.ax;
      end;
  end;

{$ELSE} {Real mode}

function MRCICompress(var NormalBuffer;
                      var CompressBuffer;
                          NormalSize : word;
                      var CompressSize : word) : word;
  var
    Request : TMRCIRequest;
  begin
    if not MRCIIsThere then
      MRCICompress := mrciErrNotSupported
    else if (NormalSize > MRCIInfo^.miMaxBuffer) then
      MRCICompress := mrciErrBadBufferSize
    else if (InDOSFlag^ <> 0) then
      MRCICompress := mrciErrBusy
    else
      begin
        FillChar(Request, sizeof(Request), 0);
        with Request do
          begin
            mrSrc := @NormalBuffer;
            mrSrcSize := NormalSize;
            mrDest := @CompressBuffer;
            if (NormalSize < MinCompressDestSize) then
                 mrDestSize := MinCompressDestSize
            else mrDestSize := NormalSize;
            mrGranularity := 512; {ie maximum compression}
          end;
        asm
          push ds
          mov ax, mrciCapCompress
          xor cx, cx
          les bx, MRCIInfo
          mov si, ss;  mov ds, si
          lea si, Request

          push bp
          push ax
          mov ax, $8001
          int $2A
          pop ax

          call DWORD PTR es:[bx].TMRCIInfo.miEntryPoint

          push ax
          mov ax, $8101
          int $2A
          pop ax
          pop bp

          or ax, ax
          jnz @@Exit
          les di, CompressSize
          mov bx, [si].TMRCIRequest.mrDestSize
          mov es:[di], bx
        @@Exit:
          pop ds
          mov @Result, ax
        end;
      end;
  end;

function MRCIDecompress(var NormalBuffer;
                        var CompressBuffer;
                        var NormalSize : word;
                            CompressSize : word) : word;
  var
    Request : TMRCIRequest;
  begin
    if not MRCIIsThere then
      MRCIDecompress := mrciErrNotSupported
    else if (NormalSize > MRCIInfo^.miMaxBuffer) then
      MRCIDecompress := mrciErrBadBufferSize
    else if (InDOSFlag^ <> 0) then
      MRCIDecompress := mrciErrBusy
    else
      begin
        FillChar(Request, sizeof(Request), 0);
        with Request do
          begin
            mrSrc := @CompressBuffer;
            mrDest := @NormalBuffer;
            mrDestSize := NormalSize;
          end;
        asm
          push ds
          mov ax, mrciCapDecompress
          xor cx, cx
          les bx, MRCIInfo
          mov si, ss;  mov ds, si
          lea si, Request

          push bp
          push ax
          mov ax, $8001
          int $2A
          pop ax

          call DWORD PTR es:[bx].TMRCIInfo.miEntryPoint

          push ax
          mov ax, $8101
          int $2A
          pop ax
          pop bp

          or ax, ax
          jnz @@Exit
          les di, NormalSize
          mov bx, [si].TMRCIRequest.mrDestSize
          mov es:[di], bx
        @@Exit:
          pop ds
          mov @Result, ax
        end;
      end;
  end;
{$ENDIF}

{$IFDEF ProtMode}
{=CleanUpDPMI=========================================================
Deallocates the selector for the MRCI info block and the InDOS flag
pointer, and deallocates the DOS memory.
08Feb94 JMB
======================================================================}
procedure CleanUpDPMI; far;
  var
    PP : pointer;
  begin
    ExitProc := ExitSave;
    if (MRCIInfo <> nil) then
      if FreeMappedDPMIPtr(MRCIInfo) then
        {nothing};
    if (InDOSFlag <> nil) then
      if FreeMappedDPMIPtr(InDOSFlag) then
        {nothing};
    if (ProtBuffer <> nil) then
      if DOSFreeMem(ProtBuffer) then
        {nothing};
  end;
{$ENDIF}

{=Initialisation======================================================
Checks for presence of MRCI server, gets the MRCI information block.
08Feb94 JMB
======================================================================}
const
  cMR = $4D52; {ie 'MR' back-words}
  cCI = $4349; {ie 'CI' back-words}
  cIC = $4943; {ie 'IC' back-words}
  cRM = $524D; {ie 'RM' back-words}
var
  {$IFDEF ProtMode}
  Regs : TRegisters;
  {$ELSE}
  Regs : Registers;
  {$ENDIF}
begin
  MRCIIsThere := false;
  {$IFDEF ProtMode}
  InDOSFlag := nil;
  ProtBuffer := nil;
  {$ENDIF}
  FillChar(Regs, sizeof(Regs), 0);
  with Regs do
    begin
      ax := $4A12;
      cx := cMR;
      dx := cCI;
      {$IFDEF ProtMode}
      RealIntr($2F, Regs);
      {$ELSE}
      Intr($2F, Regs);
      {$ENDIF}
      if (cx <> cIC) or (dx <> cRM) then
        begin
          FillChar(Regs, sizeof(Regs), 0);
          ax := $B001;
          cx := cMR;
          dx := cCI;
          {$IFDEF ProtMode}
          RealIntr($1A, Regs);
          {$ELSE}
          Intr($1A, Regs);
          {$ENDIF}
        end;
      if (cx = cIC) and (dx = cRM) then
        begin
          MRCIIsThere := true;
          {$IFDEF ProtMode}
          MRCIInfoReal := Ptr(es, di);
          if GetMappedDPMIPtr(MRCIInfo, MRCIInfoReal, sizeof(TMRCIInfo)) then
            begin
              ExitSave := ExitProc;
              ExitProc := @CleanUpDPMI;
            end
          else
            begin
              MRCIInfo := nil;
              MRCIIsThere := false;
            end;
          {$ELSE}
          MRCIInfo := Ptr(es, di);
          {$ENDIF}
        end;
    end;

  if MRCIIsThere then
    begin
      FillChar(Regs, sizeof(Regs), 0);
      Regs.ah := $34;
      {$IFDEF ProtMode}
      RealIntr($21, Regs);
      if not GetMappedDPMIPtr(InDOSFlag, Ptr(Regs.es, Regs.bx),
                              sizeof(longint)) then
        begin
          InDOSFlag := nil;
          MRCIIsThere := false;
        end;
      {$ELSE}
      Intr($21, Regs);
      InDOSFlag := Ptr(Regs.es, Regs.bx);
      {$ENDIF}
    end;

  {$IFDEF ProtMode}
  if MRCIIsThere then
    if not DOSGetMem(RealBuffer, ProtBuffer,
                     MaxDPMIBufferSize*2 + sizeof(TMRCIRequest)) then
      begin
        ProtBuffer := nil;
        MRCIIsThere := false; {well, it might as well be, we can't use it}
      end;
  {$ENDIF}
end.
